/*****************************************************************************/
/*
                                CgiLib.c

For C Language scripts these functions provide a number of common CGI
activities as a simple '#include'able code module.  One objective is the
reasonably transparent, core support for WASD CGI, CGIplus, "vanilla" CGI (e.g.
Netscape FastTrack), and OSU DECnet-based scripting.

This library is designed to have the source code included directly into the
user code (rather than create an object library) so that some control may be
exercised over the compile-time environment.  See the macros in CGILIB.H, and
instances of WASD scripting use.

The following is a list of available functions (see descriptions below for
further information).  Functions beginning 'CgiLib__' are for internal CGILIB
use only, and should not directly be called by script code.

  char* CgiVar (char *VarName);
  char* CgiLibVar (char *VarName);
  char* CgiLibVarNull (char *VarName);

  CgiLibOsuInit (int argc, char *argv[]);

  CgiLibOsuStdoutCgi ();
  CgiLibOsuStdoutRaw ();

  int CgiLibUrlDecode (char *String);

  int CgiLibUrlEncode (char *PlainString,
                      int NumberOfChars,
                      char *EncodedString,
                      int SizeOfEncodedString);

  int CgiLibHtmlEscape (char *PlainString,
                       int NumberOfChars,
                       char *EscapedString,
                       int SizeOfEscapedString);

  char* CgiLibHttpStatusCodeText (int StatusCode);

  char* CgiLibReadRequestBody (char **BodyPtrPtr, int *BodyLength);
  int CgiLibFormRequestBody (char *BodyPtr, int BodyLength);


CGI VARIABLES ... CGILIB_CGIVAR_SUPPORT macro
-------------
In particular, this library provides a simple, uniform mechanism for access to
CGI variable values transparently in the WASD CGI and CGIplus environments, as
well as OSU (via the request dialog phase) and "vanilla" CGI environments
(e.g. Netscape FastTrack).  CGI variables names are always assumed to be in
upper-case and to be preceded by "WWW_".  Will automatically detect and adjust
behaviour according to whether the CGI variables are actually prefixed with
"WWW_" (as with WASD, OSU and Purveyor) or not (as with Netscape FastTrack).

Simply call the CgiLibVar() function (or CgiVar() for backward-compatibility)
with the CGI variable name to be accessed and if that CGI variable exists a
pointer to the value will be returned.  Non-existant variables return
CGILIB_NONE (see below), which is by default an empty string ("").  The values
pointed to should NOT be modified in any way (copy the value if this is
required).

  char* CgiVar (char *VarName);
  char* CgiLibVar (char *VarName);

The following macro can be used to determine whether an empty string is
returned for a variables that does not exists (the default) or a NULL pointer.

  #define CGILIB_NONE NULL

Function CgiLibVarNull() will always return NULL if the CGI variable does not
exist, or a pointer to the variable string value.  This may be used to
differentiate between a CGI variable containing an empty value and not actually
existing at all.

  char* CgiLibVarNull (char *VarName);


CGIPLUS ... CGILIB_CGIPLUS_SUPPORT macro
-------
In the CGIplus environment supplying an empty variable name (i.e. "") will
cause the function to block until the arrival of a request, upon which it just
returns a CGILIB_NONE pointer, allowing start-of-request synchronization.  This
also resets in preparation for reading the CGI data stream of the next request,
an alternative to calling with a NULL parameter at the end of request
processing (see below).

Supplying a wildcard variable name (i.e. "*") returns each CGIplus variable
'name=value' string in turn, with no more being indicated by a NULL pointer.


OSU ENVIRONMENT ... CGILIB_OSU_SUPPORT macro
---------------
The OSU (DECthreads) HTTP server DECnet-based scripting environment is
supported semi-transparently.  When OSU is detected the CgiLibOsuInit()
initializes the OSU request information (by engaging in the dialog phase)
storing CGI-like variables for later access by CgiLibVar().  It must be
supplied with command line arguments.

  CgiLibOsuInit (int argc, char *argv[]);

Supplying a wildcard variable name (i.e. "*") returns each OSU variable
'name=value' string in turn, with no more being indicated by a NULL pointer.

The function CgiLibStdoutRaw() reopen()s <stdout> to NET_LINK.

  CgiLibOsuStdoutRaw ();

Script responses are designed to be returned in OSU "raw" mode; the script
taking care of the full response header and correctly carriage-controlled data
stream, text or binary!!  The script standard output stream is reopened in
binary mode (no translation of '\n') with a maximum record size set to 4096
bytes (failure to set this results in the OSU server reporting
%SYSTEM-F-DATAOVERUN errors).  The binary output is enclosed by suitably
fflushed() "<DNETRAW>" and "</DNETRAW>" control tags (the flush make them
individual records as required by OSU).  If in debug the script output is
placed into "text" mode.

The function CgiLibStdoutCgi() reopen()s <stdout> to NET_LINK.

  CgiLibOsuStdoutCgi ();

Script responses are designed to be returned in OSU "CGI" mode;  very similar
to other implementations where the output begins with a CGI-compliant header.
The script standard output stream is reopened in binary mode (no translation of
'\n') with a maximum record size set to 4096. The script output is enclosed by
suitably fflushed() "<DNETCGI>" and "</DNETCGI>" control tags (the flush make
them individual records as required by OSU).  If in debug the script output is
placed into "text" mode.


URL-DECODE ... CGILIB_GENERAL_SUPPORT macro
----------
Decode a URL-encoded string (i.e. string containing "+" substituted for spaces
and other forbidden characters encoded using the "%xx" hexadecimal scheme). 
Resultant string is always the same size or smaller so it can be done in-situ! 
Returns the size of the resultant string.

  int CgiLibUrlDecode (char *String);


URL-ENCODE ... CGILIB_GENERAL_SUPPORT macro
----------
URL-encode all non-alpha-numeric characters. If 'EncodedString' is NULL
sufficient memory will be dynamically allocated to hold the encoded string. 
For fixed size 'EncodedString' it will not overflow the supplied string but
does not return any indication it reached the limit.  Returns the length of the
encoded string or a pointer to the encoded string (which should be cast to
(char*) by the calling routine).
 
  int CgiLibUrlEncode (char *PlainString,
                       int NumberOfChars,
                       char *EncodedString,
                       int SizeOfEncodedString);

If 'NumberOfChars' is -1 all characters up to the null are copied.  If
'SizeOfEncodedString' is -1 the target storage if considered "big enough".


HTML-ESCAPE ... CGILIB_GENERAL_SUPPORT macro
-----------
Escape plain-text characters forbidden to occur in HTML documents (i.e. '<',
'>' and '&').  If 'EscapedString' is NULL sufficient memory will be dynamically
allocated to hold the encoded string.  For fixed size 'EscapedString' it will
not overflow the supplied string but does not return any indication it reached
the limit.  Returns the length of the escaped string or a pointer to the
encoded string (which should be cast to (char*) by the calling routine).

  int CgiLibHtmlEscape (char *PlainString,
                        int NumberOfChars,
                        char *EscapedString,
                        int SizeOfEscapedString);

If 'NumberOfChars' is -1 all characters up to the null are copied.  If
'SizeOfEscapedString' is -1 the target storage if considered "big enough".


STATUS MEANING ... CGILIB_GENERAL_SUPPORT macro
--------------
Returns a string to a "standard" status code meaning, e.g. 200 "Success".

   char* CgiLibHttpStatusCodeText (int StatusCode);


PROCESSING POSTED REQUESTS ... CGILIB_POST_SUPPORT macro
--------------------------
The following function will read a POSTed request body into a single array of
char (doesn't matter whether it's text or binary).  Returns NULL if it's not a
POST, a pointer if it is.  The caller must free() this memory when finished
with it.

  char* CgiLibReadRequestBody (char **BodyPtrPtr, int *BodyLength);

This works with WASD, "vanilla" CGI and OSU environments.  With OSU it _must_
be called before initializing output (with CgiLibOsuStdoutCgi/Raw()) which
takes the script out of dialog and into output phase.

The next function takes a "application/x-www-form-urlencoded" body (which could
have been read with the above function) and decodes the fields inserting them
into the list of CGI variables as if they were "WWW_FORM_" values processed
from the query string.  The values may then be accessed as if standard CGI
variables using CgiLibVar("WWW_FORM_...").

  int CgiLibFormRequestBody (char *BodyPtr, int BodyLength);


DEBUG STATEMENTS
----------------
Debug statements could be tuned on or off by include a 'Debug' boolean in a
macro such as the following (see WASD scripting examples for further example).
By default it is set to zero and so debug statements should be optimized out.

  #define CGILIB_DEBUG (Debug)


INCLUDING IN YOUR CODE
----------------------
This library is designed to have the source code included directly into the
user code (rather than create an object library for instance) so that some
control may be exercised over the compile-time environment.
See the macros in CGILIB.H.

Include the CGILIB.C function code from the standard WASD location
(HT_ROOT:[SRC.MISC]) if CGILIB_INCLUDE has not been defined. If #defined
then the value must be the #include location.

A number of "boolean" macros control which code is included and which is not. 
These will make final executables more compact if not all functionality is
required.  Making CGILIB_ALL_SUPPORT true includes all available functionality.

The following example provides a suggested method for doing this.

  #define CGILIB_CGIVAR_SUPPORT  1
  #define CGILIB_CGIPLUS_SUPPORT 1
  #define CGILIB_OSU_SUPPORT     0
  #define CGILIB_GENERAL_SUPPORT 1
  #define CGILIB_POST_SUPPORT    0

  #ifndef CGILIB_INCLUDE_H
  #   include "/ht_root/src/misc/cgilib.h"
  #   include "/ht_root/src/misc/cgilib.c"
  #else
  #   include CGILIB_INCLUDE_H
  #   include CGILIB_INCLUDE_C
  #endif


COPYRIGHT
---------
Copyright (c) 1999 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update SoftwareID in header file as well!)
---------------
24-APR-1999  MGD  v1.0.0, initial '#include'able code module
*/
/*****************************************************************************/

/* standard C headers */
#include <ctype.h>
#include <errno.h>
#include <file.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* VMS headers */
#include <stsdef.h>
#include <ssdef.h>

/* module header */
#ifndef CGILIB_H_LOADED
#   include "/ht_root/src/misc/cgilib.h"
#endif

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  Returns the size of the resultant string.
*/

#if CGILIB_GENERAL_SUPPORT
#   define CGILIB_URL_DECODE 1
#endif
#if CGILIB_CGIVAR_NON_WWW
#   define CGILIB_URL_DECODE 1
#endif
#if CGILIB_POST_SUPPORT
#   define CGILIB_URL_DECODE 1
#endif
#ifndef CGILIB_URL_DECODE
#   define CGILIB_URL_DECODE 0
#endif

#if CGILIB_URL_DECODE

int CgiLibUrlDecode (char *String)

{
   register char  c;
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLibUrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            c = 0;
            if (*cptr >= '0' && *cptr <= '9')
               c = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               c += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = c;

            continue;

         case '\r' :
         case '\n' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

#endif /* CGILIB_URL_DECODE */

/****************************************************************************/
/*
URL-encode all non-alpha-numeric characters. If 'EncodedString' is NULL
sufficient memory will be dynamically allocated to hold the encoded string. 
For fixed size 'EncodedString' it will not overflow the supplied string but
does not return any indication it reached the limit.  Returns the length of the
encoded string or a pointer to the encoded string (which should be cast to
(char*) by the calling routine).
*/ 
 
#if CGILIB_GENERAL_SUPPORT

int CgiLibUrlEncode
(
char *PlainString,
int NumberOfChars,
char *EncodedString,
int SizeOfEncodedString
)
{
   static char  HexDigits [] = "0123456789abcdef";

   register int  ccnt;
   register char  *cptr, *sptr, *zptr;

   int  CurrentLength,
        EncodedSize;
   char  *EncodedPtr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG)
      fprintf (stdout, "CgiLibUrlEncode() |%s| %d\n",
               PlainString, NumberOfChars);

   if (!*(cptr = PlainString))
   {
      if (EncodedString == NULL)
         return ((int)"");
      else
         return (0);
   }

   if (EncodedString == NULL)
   {
      sptr = zptr = EncodedPtr = NULL;
      EncodedSize = 0;
   }
   else
   {
      sptr = EncodedPtr = EncodedString;
      if (SizeOfEncodedString == -1)
         zptr = (char*)INT_MAX;
      else
         zptr = sptr + SizeOfEncodedString - 1;
   }

   if (NumberOfChars == -1)
      ccnt = INT_MAX;
   else
      ccnt = NumberOfChars;

   while (*cptr && ccnt)
   {
      if (EncodedString == NULL && sptr >= zptr)
      {
         /* out of storage (or first use) get more (or some) memory */
         CurrentLength = sptr - EncodedPtr;
         if ((EncodedPtr =
              realloc (EncodedPtr, EncodedSize+CGILIB_ENCODED_CHUNK)) == NULL)
            exit (vaxc$errno);
         EncodedSize += CGILIB_ENCODED_CHUNK;
         /* recalculate pointers */
         sptr = EncodedPtr + CurrentLength;
         zptr = EncodedPtr + EncodedSize - 1;
      }
      else
      if (sptr >= zptr)
         break;

      if (isalnum(*cptr) ||
          *cptr == '/' ||
          *cptr == '-' ||
          *cptr == '_' ||
          *cptr == '.')
      {
         *sptr++ = *cptr++;
         ccnt--;
         continue;
      }

      if (sptr < zptr) *sptr++ = '%';
      if (sptr < zptr) *sptr++ = HexDigits[(*cptr & 0xf0) >> 4];
      if (sptr < zptr) *sptr++ = HexDigits[*cptr & 0x0f];
      cptr++;
      ccnt--;
   }
   *sptr = '\0';
   if (CGILIB_DEBUG) fprintf (stdout, "%d |%s|\n", sptr-EncodedPtr, EncodedPtr);

   if (EncodedString == NULL)
      return ((int)EncodedPtr);
   else
      return (sptr - EncodedPtr);
}
 
#endif /* CGILIB_GENERAL_SUPPORT */

/****************************************************************************/
/*
Escape plain-text characters forbidden to occur in HTML documents (i.e. '<',
'>' and '&').  If 'EscapedString' is NULL sufficient memory will be dynamically
allocated to hold the encoded string.  For fixed size 'EscapedString' it will
not overflow the supplied string but does not return any indication it reached
the limit.  Returns the length of the escaped string or a pointer to the
encoded string (which should be cast to (char*) by the calling routine).
*/ 
 
#if CGILIB_GENERAL_SUPPORT

int CgiLibHtmlEscape
(
char *PlainString,
int NumberOfChars,
char *EscapedString,
int SizeOfEscapedString
)
{
   register int  ccnt;
   register char  *cptr, *sptr, *zptr;

   int  CurrentLength,
        EscapedSize;
   char  *EscapedPtr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG)
      fprintf (stdout, "CgiLibHtmlEscape() |%s| %d\n",
               PlainString, NumberOfChars);

   if (!*(cptr = PlainString))
   {
      if (EscapedString == NULL)
         return ((int)"");
      else
         return (0);
   }

   if (EscapedString == NULL)
   {
      sptr = zptr = EscapedPtr = NULL;
      EscapedSize = 0;
   }
   else
   {
      sptr = EscapedPtr = EscapedString;
      if (SizeOfEscapedString == -1)
         zptr = (char*)INT_MAX;
      else
         zptr = sptr + SizeOfEscapedString - 1;
   }

   if (NumberOfChars == -1)
      ccnt = INT_MAX;
   else
      ccnt = NumberOfChars;

   while (*cptr && ccnt)
   {
      /* 'sptr+6' is the "worst-case" character substitution */
      if (EscapedString == NULL && sptr+6 >= zptr)
      {
         /* out of storage (or first use) get more (or some) memory */
         CurrentLength = sptr - EscapedPtr;
         if ((EscapedPtr =
              realloc (EscapedPtr, EscapedSize+CGILIB_ESCAPED_CHUNK)) == NULL)
            exit (vaxc$errno);
         EscapedSize += CGILIB_ESCAPED_CHUNK;
         /* recalculate pointers */
         sptr = EscapedPtr + CurrentLength;
         zptr = EscapedPtr + EscapedSize - 1;
      }
      else
      if (sptr+6 >= zptr)
         break;

      switch (*cptr)
      {
         case '<' :
            memcpy (sptr, "&lt;", 4); sptr += 4; cptr++; ccnt--; continue;
         case '>' :
            memcpy (sptr, "&gt;", 4); sptr += 4; cptr++; ccnt--; continue;
         case '&' :
            memcpy (sptr, "&amp;", 5); sptr += 5; cptr++; ccnt--; continue;
         case '"' :
            memcpy (sptr, "&quot;", 6); sptr += 6; cptr++; ccnt--; continue;
         default :
            *sptr++ = *cptr++;
            ccnt--;
       }
   }
   *sptr = '\0';
   if (CGILIB_DEBUG) fprintf (stdout, "%d |%s|\n", sptr-EscapedPtr, EscapedPtr);

   if (EscapedString == NULL)
      return ((int)EscapedPtr);
   else
      return (sptr - EscapedPtr);
}
 
#endif /* CGILIB_GENERAL_SUPPORT */

/*****************************************************************************/
/*
Return the a pointer to abbreviated meaning of the supplied HTTP status code.
These are typical of those included on the response header status line.
*/

#if CGILIB_GENERAL_SUPPORT

char* CgiLibHttpStatusCodeText (int StatusCode)

{
   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "HttpStatusCodeText() %d\n", StatusCode);

   switch (StatusCode)
   {
      case 000 : return ("Script Error!");
      case 200 : return ("Success");
      case 201 : return ("Created");
      case 202 : return ("Accepted");
      case 203 : return ("No content");
      case 301 : return ("Moved permanently");
      case 302 : return ("Moved temporarily");
      case 304 : return ("Not modified");
      case 400 : return ("Bad request");
      case 401 : return ("Authorization required");
      case 403 : return ("Forbidden");
      case 404 : return ("Not found");
      case 500 : return ("Internal error");
      case 501 : return ("Not implemented");
      case 502 : return ("Bad gateway");
      case 503 : return ("Service unavailable");
      default :  return ("Unknown code!");
   }
}

#endif /* CGILIB_GENERAL_SUPPORT */

/*****************************************************************************/
/*
Get a CGI variable from a DCL symbol or from the CGIplus stream.  Returns a
pointer to the string value if the CGI variables exists or a pointer to
'NonePtr' if not.  Adjusts behaviour according to CGI or CGIplus environments
and to whether symbols are prefixd with "WWW_" or not.
*/

#ifdef CGILIB_CGIVAR_SUPPORT

char* CgiLib__GetVar
(
char *VarName,
char *NonePtr
)
{
   static int  CGIplusEnvironment = 0,
               InitializeEnvironment = 1,
               LoadVariables = 1,
               OsuEnvironment = 0,
               WWWprefix = 0;

   static char  WwwVarName [256] = "WWW_";

   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__GetVar() |%s|\n", VarName);

   if (InitializeEnvironment)
   {
      InitializeEnvironment = 0;
      WWWprefix = (getenv ("WWW_GATEWAY_INTERFACE") != NULL);

#if CGILIB_CGIPLUS_SUPPORT
      CGIplusEnvironment = (getenv ("CGIPLUSEOF") != NULL);
#endif

#if CGILIB_OSU_SUPPORT
      OsuEnvironment = (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL);
#endif
   }

#if CGILIB_CGIVAR_WWW

   if (VarName != NULL &&
       VarName[0] &&
       VarName[0] != '*' &&
       memcmp (VarName, "WWW_", 4))
   {
      /* variable name does not begin with "WWW_", make one that does */
      strncpy (WwwVarName+4, VarName, sizeof(WwwVarName)-5);
      (VarName = WwwVarName)[sizeof(WwwVarName)-1] = '\0';
   }

#endif /* CGILIB_CGIVAR_WWW */

   if (!CGIplusEnvironment)
   {
      /***************************/
      /* non-CGIplus environment */
      /***************************/

      if (VarName[0] == '*' && !OsuEnvironment) return (NULL);

#if CGILIB_OSU_SUPPORT
      if (OsuEnvironment)
         return (CgiLib__VarList (VarName, NULL, NonePtr));
#endif

      if (WWWprefix)
      {
         /* WASD, probably Purveyor */
         if ((cptr = getenv(VarName)) == NULL)
         {
            if (CGILIB_DEBUG) fprintf (stdout, "|%s|=CGILIB_NONE\n", VarName);
            return (NonePtr);
         }
         else
         {
            if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, cptr);
            return (cptr);
         }
      }

      /* at least Netscape FastTrack */
      if (!memcmp (VarName+4, "KEY_", 4))
         return (CgiLib__GetKey (VarName+4, NonePtr));

      if (!memcmp (VarName+4, "FORM_", 5))
         return (CgiLib__GetForm (VarName+9, NonePtr));

      if (!strcmp (VarName+4, "PATH_TRANSLATED"))
         return (CgiLib__GetPathTranslated (NonePtr));

      if ((cptr = getenv(VarName+4)) == NULL)
      {
         if (CGILIB_DEBUG) fprintf (stdout, "|%s|=CGILIB_NONE\n", VarName);
         return (NonePtr);
      }
      else
      {
         if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, cptr);
         return (cptr);
      }
   }

#if CGILIB_CGIPLUS_SUPPORT

   /***********************/
   /* CGIplus environment */
   /***********************/

   if (VarName == NULL || !VarName[0])
   {
      /* end of request, reinitialize */
      LoadVariables = 1;
      CgiLib__VarList (NULL, NULL, NULL);

      /* if just initializing then return now */
      if (VarName == NULL) return (NULL);
   }

   if (LoadVariables)
   {
      /* read lines containing CGIplus variables from <CGIPLUSIN> */
      static FILE  *CgiPlusIn = NULL;
      char  Line [CGILIB_CGIPLUS_LINE_SIZE];

      if (CgiPlusIn == NULL)
      {
         if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
            exit (vaxc$errno);
      }

      LoadVariables = 0;

      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", Line);
         if (Line[0] == '\n') break;
         if (!isalpha(Line[0])) continue;
#if CGILIB_CGIVAR_WWW
         /* include the leading "WWW_" */
         CgiLib__VarList (Line, "", NULL);
#else
         /* variables will never be prefixed by "WWW_", ignore it */
         CgiLib__VarList (Line+4, "", NULL);
#endif /* CGILIB_CGIVAR_WWW */
      }
   }

   /* if waiting for next request return now it's arrived */
   if (!VarName[0]) return (NULL);

   /* get a CGIplus variable */
   return (CgiLib__VarList (VarName, NULL, NonePtr));

#endif /* CGILIB_CGIPLUS_SUPPORT */
}

#endif /* CGILIB_CGIVAR_SUPPORT */

/*****************************************************************************/
/*
Set or get a CGI variable.  Used to store CGIplus variables (read from
CGIPLUSIN stream), OSU variables (created individually during dialog phase) and
the pseudo-CGI-variables from a form-URL-encoded body.

Implemented using a simple list in allocated memory.  Each entry comprises a
total-length-of-entry short, followed by a length-of-variable-name short,
followed by a null-terminated string comprising the variable name, an equate
symbol, and the variable value string.

If 'VarValue' is NULL then get 'VarName', returning a pointer to it's string
value, or if not found to CGILIB_NONE (usually "").  Successive calls with
'VarName' set to "*" (via CgiLibVar() of course) returns each of the list's
entries in turn, a NULL indicating end-of-list.  The list can be reset (for
CGIplus use) with 'VarName' equal to NULL.  If 'VarValue' is not NULL then
store the value against the name (duplicated variables names result in
duplicated entires, not overwriting!)
*/

#if CGILIB_CGIPLUS_SUPPORT
#define CGILIB_VARLIST 1
#endif
#if CGILIB_OSU_SUPPORT
#define CGILIB_VARLIST 1
#endif
#if CGILIB_BODY_SUPPORT
#define CGILIB_VARLIST 1
#endif
#ifndef CGILIB_VARLIST
#define CGILIB_VARLIST 0
#endif

#ifdef CGILIB_VARLIST

char* CgiLib__VarList
(
char *VarName,
char *VarValue,
char *NonePtr
)
{
   static int  ListNextOffset = 0,
               ListSize = 0;

   static char  *NextPtr = NULL,
                *ListPtr = NULL;

   register char  *cptr, *sptr, *zptr;

   int  VarNameLength;
   unsigned short  Length;
   char  *LengthPtr,
         *VarNamePtr,
         *VarNameLengthPtr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG)
      fprintf (stdout, "CgiLib__VarList() |%s|%s|\n",
               (VarName == NULL ? "(null)" : VarName),
               (VarValue == NULL ? "(null)" : VarValue));

   if (VarName == NULL)
   {
      /* initialize */
      NextPtr = NULL;
      ListNextOffset = 0;
      return (NULL);
   }

   if (VarValue != NULL)
   {
      /******************/
      /* add a variable */
      /******************/

      if (ListPtr == NULL)
      {
         /* always allow extra space (allows sloppier/faster code ;^) */
         ListPtr = realloc (ListPtr, CGILIB_VARLIST_CHUNK+16);
         if (ListPtr == NULL) exit (vaxc$errno);
         ListSize = CGILIB_VARLIST_CHUNK;
      }

      /* point to next "[total-len][var-name-len]var-name=var-value" storage */
      sptr = LengthPtr = ListPtr + ListNextOffset;
      /* now point to where the length of the variable name is stored */
      VarNameLengthPtr = (sptr += sizeof(unsigned short));
      /* now point to the start of the "var-name=var-value" string */
      VarNamePtr = (sptr += sizeof(unsigned short));
      /* so we don't over-run the list storage */
      zptr = ListPtr + ListSize;

      for (cptr = VarName; *cptr; *sptr++ = *cptr++)
      {
         /* CGIplus, stream has "name=value\n" */
         if (*cptr == '\n') break;
         if (*cptr == '=' && VarNameLengthPtr != NULL)
         {
            /* CGIplus, store the length of the variable name */
            *((unsigned short*)VarNameLengthPtr) = sptr - VarNamePtr;
            /* indicate we've already done this on the first '=' */
            VarNameLengthPtr = NULL;
         }
         if (sptr < zptr) continue;
         /* out of storage, get more memory */
         Length = sptr - LengthPtr;
         ListPtr = realloc (ListPtr, ListSize+CGILIB_VARLIST_CHUNK+16);
         if (ListPtr == NULL) exit (vaxc$errno);
         ListSize += CGILIB_VARLIST_CHUNK;
         /* recalculate pointers */
         LengthPtr = ListPtr + ListNextOffset;
         sptr = LengthPtr + Length;
         zptr = ListPtr + ListSize - 2;
      }
      if (*cptr != '\n')
      {
         /* store the length of the variable name */
         *((unsigned short*)VarNameLengthPtr) = sptr - VarNamePtr;
         *sptr++ = '=';
         /* add the variable value */
         for (cptr = VarValue; *cptr; *sptr++ = *cptr++)
         {
            if (sptr < zptr) continue;
            /* out of storage get more memory */
            Length = sptr - LengthPtr;
            ListPtr = realloc (ListPtr, ListSize+CGILIB_VARLIST_CHUNK+16);
            if (ListPtr == NULL) exit (vaxc$errno);
            ListSize += CGILIB_VARLIST_CHUNK;
            /* recalculate pointers */
            LengthPtr = ListPtr + ListNextOffset;
            sptr = LengthPtr + Length;
            zptr = ListPtr + ListSize;
         }
      }
      *sptr++ = '\0';

      /* insert the length of this name=value pair immediately before it */
      *((unsigned short*)LengthPtr) = Length = sptr - LengthPtr;

      /* adjust the value of the current variable storage space used */
      ListNextOffset += Length;

      /* ensure the list is terminated with a zero length entry */
      *(unsigned short*)(ListPtr+ListNextOffset) = 0;

      if (CGILIB_DEBUG)
      {
         fprintf (stdout, "ListPtr: %d ListSize: %d ListNextOffset: %d\n",
                  ListPtr, ListSize, ListNextOffset);
         sptr = ListPtr;
         for (sptr = ListPtr;
              Length = *((unsigned short*)sptr);
              sptr += Length)
         {
            fprintf (stdout, "%5d %2d |%s|\n",
               Length, *((unsigned short*)(sptr+sizeof(unsigned short))),
               sptr+sizeof(unsigned short)+sizeof(unsigned short));
         }
      }

      return (NULL);
   }

   if (ListPtr == NULL) return (NonePtr);

   if (VarName[0] == '*')
   {
      /*****************************************/
      /* return each stored line one at a time */
      /*****************************************/

      if (NextPtr == NULL) NextPtr = ListPtr;
      for (;;)
      {
         if (!(Length = *((unsigned short*)(sptr = NextPtr))))
         {
            if (CGILIB_DEBUG) fprintf (stdout, "|*|=NULL\n");
            return (NextPtr = NULL);
         }
         NextPtr = sptr + Length;
         sptr += sizeof(unsigned short) + sizeof(unsigned short);
         if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", sptr);
         if (!VarName[1]) return (sptr);
      }
      return (NULL);
   }

   /*************************************************/
   /* search variable strings for required variable */
   /*************************************************/

   VarNameLength = strlen(VarName);

   for (sptr = LengthPtr = ListPtr;
        Length = *((unsigned short*)sptr);
        sptr = LengthPtr = LengthPtr + Length)
   {
      /* step over the total-length short */
      sptr += sizeof(unsigned short);
      if (*(unsigned short*)sptr != VarNameLength) continue;
      /* step over the var-name-length short */
      sptr += sizeof(unsigned short);
      /* simple comparison between supplied and in-string variable names */
      cptr = VarName;
      /** if (CGILIB_DEBUG) fprintf (stdout, "|%s|%s|\n", cptr, sptr); **/
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      if (*cptr || *sptr != '=') continue;
      /* found, return a pointer to the value */
      if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, sptr+1);
      return (sptr+1);
   }

   /* not found */
   if (CGILIB_DEBUG) fprintf (stdout, "|%s|=CGILIB_NONE\n", VarName);
   return (NonePtr);
}  

#endif /* CGILIB_VARLIST */

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) derive the
"KEY_COUNT", "KEY_1" ... "KEY_n" variables from "QUERY_STRING".
*/

#if CGILIB_CGIVAR_NON_WWW

char* CgiLib__GetKey
(
char *VarName,
char *NonePtr
)
{
   static char  *QueryStringPtr = NULL;

   register char  *aptr, *cptr, *sptr;

   int  KeyCount,
        VarKeyCount;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__GetKey() |%s|\n", VarName);

   if (!VarName[0]) return (NonePtr);

   if ((cptr = QueryStringPtr) == NULL)
      if ((QueryStringPtr = cptr = getenv ("QUERY_STRING")) == NULL)
         return (NonePtr);
   if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", QueryStringPtr);

   if (!strcmp (VarName, "KEY_COUNT"))
   {
      KeyCount = 0;
      for (sptr = cptr; *sptr; sptr++)
      {
         if (*sptr == '=') return (NonePtr);
         if (*sptr == '+') KeyCount++;
      }
      if (sptr == cptr) return ("0");
      KeyCount++;
      sprintf (aptr = calloc (1, 16), "%d", KeyCount);
      if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
      return (aptr);
   }
 
   VarKeyCount = atoi (VarName+4);
   KeyCount = 1;
   while (*cptr && VarKeyCount)
   {
      sptr = cptr;
      while (*cptr && *cptr != '+')
      {
         if (*cptr == '=') return (NonePtr);
         cptr++;
      }
      if (KeyCount == VarKeyCount)
      {
         aptr = calloc (1, cptr - sptr + 1);
         memcpy (aptr, sptr, cptr-sptr);
         aptr[cptr-sptr] = '\0';
         CgiLibUrlDecode (aptr);
         if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
         return (aptr);
      }
      if (*cptr) cptr++;
      KeyCount++;
   }

   if (CGILIB_DEBUG) fprintf (stdout, "|%s|=CGILIB_NONE\n", VarName);
   return (NonePtr);
}

#endif /* CGILIB_CGIVAR_NON_WWW */

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) derive the
"FORM_xxxxx" variables from "QUERY_STRING".
*/

#if CGILIB_CGIVAR_NON_WWW

char* CgiLib__GetForm
(
char *VarName,
char *NonePtr
)
{
   static char  *QueryStringPtr = NULL;

   register char  *aptr, *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__GetForm() |%s|\n", VarName);

   if ((cptr = QueryStringPtr) == NULL)
      if ((QueryStringPtr = cptr = getenv ("QUERY_STRING")) == NULL)
         return (NonePtr);
   if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", QueryStringPtr);

   while (*cptr)
   {
      sptr = cptr;
      while (*cptr && *cptr != '=') cptr++;
      if (!*cptr) return (NonePtr);
      *cptr = '\0';
      aptr = VarName;
      while (*aptr && *sptr && toupper(*aptr) == toupper(*sptr))
      {
         aptr++;
         sptr++;
      }
      if (*aptr || *sptr)
      {
         *cptr++ = '=';
         while (*cptr && *cptr != '&') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      *cptr++ = '=';

      sptr = cptr;
      while (*cptr && *cptr != '&') cptr++;
      aptr = calloc (1, cptr - sptr + 1);
      memcpy (aptr, sptr, cptr-sptr);
      aptr[cptr-sptr] = '\0';
      CgiLibUrlDecode (aptr);
      if (CGILIB_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
      return (aptr);
   }

   if (CGILIB_DEBUG) fprintf (stdout, "|%s|=CGILIB_NONE\n", VarName);
   return (NonePtr);
}

#endif /* CGILIB_CGIVAR_NON_WWW */

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) get the
"PATH_TRANSLATED" variable and check if it's in Unix-style or VMS-style format. 
Convert to a VMS specification as necessary.
*/

#if CGILIB_CGIVAR_NON_WWW

char* CgiLib__GetPathTranslated (char *NonePtr)
   
{
   register char  *aptr, *cptr, *sptr;

   int  Length;
   char  PathTranslated [256];

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__GetPathTranslated()\n");

   if ((aptr = cptr = getenv ("PATH_TRANSLATED")) == NULL)
      return (NonePtr);

   if (*cptr == '/')
   {
      /* unix-style specification */
      sptr = PathTranslated;
      if (*cptr == '/') cptr++;
      while (*cptr && *cptr != '/') *sptr++ = toupper(*cptr++);
      if (*cptr)
      {
         cptr++;
         *sptr++ = ':';
         *sptr++ = '[';
      }
      while (*cptr) *sptr++ = toupper(*cptr++);
      *sptr-- = '\0';
      while (sptr > PathTranslated && *sptr != '/') sptr--;
      if (*sptr == '/')
      {
         *sptr-- = ']';
         while (sptr >= PathTranslated)
         {
            if (*sptr == '/') *sptr = '.';
            if (*(unsigned long*)sptr == '....')
               memmove (sptr, sptr+1, strlen(sptr)); 
            sptr--;
         }
      }
      aptr = calloc (1, Length = strlen(PathTranslated) + 1);
      memcpy (aptr, PathTranslated, Length);
   }

   if (CGILIB_DEBUG) fprintf (stdout, "|PATH_TRANSLATED=%s|\n", aptr);
   return (aptr);
}

#endif /* CGILIB_CGIVAR_NON_WWW */

/*****************************************************************************/
/*
Initialize the OSU variable list using the request network dialog phase.
*/

#if CGILIB_OSU_SUPPORT

void CgiLibOsuInit
(
int argc,
char *argv[]
)
{
   register char  *cptr, *sptr, *zptr;

   int  ContentLengthSet,
        ContentTypeSet,
        IpAddress,
        KeyCount;
   char  PathInfo [256],
         PathTranslated [256],
         Scratch [256],
         ScriptName [256],
         SymbolName [256];

   /*********/
   /* begin */
   /*********/

   if (argc < 4)
   {
      fprintf (stdout, "%%CGILIB-E-OSU, P1=<method> P2=<path> P3=<protocol>\n");
      exit (SS$_INSFARG | STS$M_INHIB_MSG);
   }

   if (CGILIB_DEBUG)
   {
      /* let's have a look at what's available! */
      fprintf (stdout, "CgiLibOsuInit()\n----------\n");
      CgiLib__OsuQueryNetLink ("<DNETARG>", 1);
      CgiLib__OsuQueryNetLink ("<DNETARG2>", 1);
      CgiLib__OsuQueryNetLink ("<DNETHOST>", 1);
      CgiLib__OsuQueryNetLink ("<DNETID>", 1);
      CgiLib__OsuQueryNetLink ("<DNETID2>", 1);
      CgiLib__OsuQueryNetLink ("<DNETRQURL>", 1);
      CgiLib__OsuQueryNetLink ("<DNETBINDIR>", 1);
      CgiLib__OsuQueryNetLink ("<DNETPATH>", 1);
      CgiLib__OsuQueryNetLink ("<DNETHDR>", 999);
      fprintf (stdout, "----------\n");
   }

   /**********/
   /* basics */
   /**********/

   CgiLib__VarList ("WWW_GATEWAY_INTERFACE", "CGI/1.1", NULL);
   CgiLib__VarList ("WWW_REQUEST_METHOD", argv[1], NULL);
   CgiLib__VarList ("WWW_SERVER_PROTOCOL", argv[3], NULL);

   /*******************************************/
   /* SCRIPT_NAME, PATH_INFO, PATH_TRANSLATED */
   /*******************************************/

   strcpy (sptr = ScriptName, CgiLib__OsuQueryNetLink ("<DNETPATH>", 1));
   cptr = CgiLib__OsuQueryNetLink ("<DNETRQURL>", 1);
   while (*cptr && *cptr && *cptr == *sptr) { cptr++; sptr++; }
   while (*cptr && *cptr != '/' && *cptr != '?') *sptr++ = *cptr++;
   *sptr = '\0';
   CgiLibUrlDecode (ScriptName);
   CgiLib__VarList ("WWW_SCRIPT_NAME", ScriptName, NULL);

   /* continue to get the path information from the request URL */
   sptr = PathInfo;
   while (*cptr && *cptr != '?') *sptr++ = *cptr++;
   *sptr = '\0';
   CgiLibUrlDecode (PathInfo);

   if (*PathInfo && *(unsigned short*)PathInfo != '/\0')
   {
      CgiLib__OsuQueryNetLink ("<DNETXLATE>", 0);
      sptr = CgiLib__OsuQueryNetLink (PathInfo, 1);

      cptr = PathTranslated;
      if (*sptr == '/') sptr++;
      while (*sptr && *sptr != '/') *cptr++ = toupper(*sptr++);
      if (*sptr)
      {
         sptr++;
         *cptr++ = ':';
         *cptr++ = '[';
      }
      while (*sptr) *cptr++ = toupper(*sptr++);
      *cptr-- = '\0';
      while (cptr > PathTranslated && *cptr != '/') cptr--;
      if (*cptr == '/')
      {
         *cptr-- = ']';
         while (cptr >= PathTranslated)
         {
            if (*cptr == '/') *cptr = '.';
            if (*(unsigned long*)cptr == '....')
               memmove (cptr, cptr+1, strlen(cptr)); 
            cptr--;
         }
      }

      CgiLib__VarList ("WWW_PATH_INFO", PathInfo, NULL);
      CgiLib__VarList ("WWW_PATH_TRANSLATED", PathTranslated, NULL);
   }
   else
   {
      CgiLib__VarList ("WWW_PATH_INFO", "/", NULL);
      CgiLib__VarList ("WWW_PATH_TRANSLATED", "", NULL);
   }

   /***********************************/
   /* QUERY_STRING, FORM_..., KEY_... */
   /***********************************/

   cptr = CgiLib__OsuQueryNetLink ("<DNETARG>", 1);
   if (*cptr == '?') cptr++;
   CgiLib__VarList ("WWW_QUERY_STRING", cptr, NULL);

   KeyCount = 0;
   sptr = cptr;
   while (*cptr && *cptr != '=') cptr++;
   if (*cptr)
   {
      /* FORM-formatted query string */
      while (*cptr)
      {
         *cptr++ = '\0';
         sprintf (SymbolName, "WWW_FORM_%s", sptr);
         sptr = cptr;
         while (*cptr && *cptr != '&') cptr++;
         if (*cptr) *cptr++ = '\0';
         CgiLibUrlDecode (SymbolName);
         CgiLibUrlDecode (sptr);
         CgiLib__VarList (SymbolName, sptr, NULL);
         sptr = cptr;
         while (*cptr && *cptr != '=') cptr++;
      }
   }
   else
   {
      /* ISQUERY-formatted query string */
      cptr = sptr;
      while (*cptr && *cptr != '+') cptr++;
      while (*sptr)
      {
         *cptr++ = '\0';
         sprintf (SymbolName, "WWW_KEY_%d", ++KeyCount);
         CgiLibUrlDecode (sptr);
         CgiLib__VarList (SymbolName, sptr, NULL);
         sptr = cptr;
         while (*cptr && *cptr != '+') cptr++;
      }
   }
   sprintf (Scratch, "%d", KeyCount);
   CgiLib__VarList ("WWW_KEY_COUNT", Scratch, NULL);

   /**********************************************/
   /* SERVER_SOFTWARE, SERVER_NAME, SERVER_PORT, */
   /* REMOTE_HOST, REMOTE_ADDR, REMOTE_USER      */
   /**********************************************/

   cptr = sptr = CgiLib__OsuQueryNetLink ("<DNETID2>", 1);

   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';
   CgiLib__VarList ("WWW_SERVER_SOFTWARE", sptr, NULL);

   sptr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';
   CgiLib__VarList ("WWW_SERVER_NAME", sptr, NULL);

   sptr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';
   CgiLib__VarList ("WWW_SERVER_PORT", sptr, NULL);

   /* skip over client port */
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';

   /* host address */
   sptr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';
   for (zptr = sptr; *zptr && *zptr != '.'; zptr++);
   if (*zptr != '.')
   {
      IpAddress = atoi(sptr);
      sprintf (Scratch, "%d.%d.%d.%d",
               *(unsigned char*)&IpAddress,
               *((unsigned char*)(&IpAddress)+1),
               *((unsigned char*)(&IpAddress)+2),
               *((unsigned char*)(&IpAddress)+3));
      while (*cptr && !isspace(*cptr)) cptr++;
      sptr = Scratch;
   }
   CgiLib__VarList ("WWW_REMOTE_ADDR", sptr, NULL);
   CgiLib__VarList ("WWW_REMOTE_HOST", sptr, NULL);

   sptr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr) *cptr++ = '\0';
   CgiLib__VarList ("WWW_REMOTE_USER", sptr, NULL);
   sptr = cptr;

   sptr = cptr;
   while (*cptr && !isspace(*cptr)) cptr++;
   if (*cptr)
   {
      /* DNS-resolved host name */
      *cptr++ = '\0';
      CgiLib__VarList ("WWW_REMOTE_HOST", sptr, NULL);
   }

   /******************************************/
   /* CONTENT_LENGTH, CONTENT_TYPE, HTTP_... */
   /******************************************/

   ContentLengthSet = ContentTypeSet = 0;

   cptr = sptr = CgiLib__OsuQueryNetLink ("<DNETHDR>", 999);
   while (*cptr)
   {
      while (*cptr && *cptr != ':')
      {
         if (!isalnum(*cptr)) *cptr = '_';
         *cptr = toupper(*cptr++);
      }
      if (*cptr)
      {
         *cptr++ = '\0';
         sprintf (SymbolName, "WWW_HTTP_%s", sptr);
         while (*cptr && isspace(*cptr) && *cptr != '\n') *cptr++;
         sptr = cptr;
         while (*cptr && *cptr != '\n') *cptr++;
         if (*cptr) *cptr++ = '\0';
         CgiLibUrlDecode (SymbolName);
         CgiLibUrlDecode (sptr);
         if (!strcmp (SymbolName+9, "CONTENT_TYPE"))
         {
            ContentTypeSet = 1;
            CgiLib__VarList ("WWW_CONTENT_TYPE", sptr, NULL);
         }
         else
         if (!strcmp (SymbolName+9, "CONTENT_LENGTH"))
         {
            ContentLengthSet = 1;
            CgiLib__VarList ("WWW_CONTENT_LENGTH", sptr, NULL);
         }
         else
            CgiLib__VarList (SymbolName, sptr, NULL);
         sptr = cptr;
      }
   }

   if (!ContentTypeSet)
      CgiLib__VarList ("WWW_CONTENT_TYPE", "", NULL);

   if (!ContentLengthSet)
      CgiLib__VarList ("WWW_CONTENT_LENGTH", "0", NULL);
}  

#endif /* CGILIB_OSU_SUPPORT */

/*****************************************************************************/
/*
Reopen <stdout> to NET_LINK: in binary mode.  This is much more efficient as the
CRTL buffers individual prints into a stream.  Output an appropriate
output-phase tag and establish an exit function to complete the output dialog.
*/

#if CGILIB_OSU_SUPPORT

void CgiLibOsuStdoutCgi ()

{
   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLibOsuStdoutCgi()\n");

   /* reopen so that '\r' and '\n' are not filtered, maximum record size */
   if ((stdout =
       freopen ("NET_LINK:", "w", stdout, "ctx=bin", "mrs=4096")) == NULL)
      exit (vaxc$errno);
 
   if (CGILIB_DEBUG)
   {
      fputs ("<DNETTEXT>", stdout);
      fflush(stdout);
      fputs ("200 Debug", stdout);
   }
   else
      fputs ("<DNETCGI>", stdout);
   fflush(stdout);

   atexit (&CgiLib__OsuExitCgi);
}  

void CgiLib__OsuExitCgi ()

{
   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__OsuExitCgi()\n");

   fflush(stdout);
   if (CGILIB_DEBUG)
      fputs ("</DNETTEXT>", stdout);
   else
      fputs ("</DNETCGI>", stdout);
   fflush(stdout);
}  

#endif /* CGILIB_OSU_SUPPORT */

/*****************************************************************************/
/*
Reopen <stdout> to NET_LINK: in binary mode.  This is much more efficient as the
CRTL buffers individual prints into a stream.  Output an appropriate
output-phase tag and establish an exit function to complete the output dialog.
*/

#if CGILIB_OSU_SUPPORT

void CgiLibOsuStdoutRaw ()

{
   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLibOsuStdoutRaw()\n");

   /* reopen so that '\r' and '\n' are not filtered, maximum record size */
   if ((stdout =
       freopen ("NET_LINK:", "w", stdout, "ctx=bin", "mrs=4096")) == NULL)
      exit (vaxc$errno);

   if (CGILIB_DEBUG)
   {
      fputs ("<DNETTEXT>", stdout);
      fflush(stdout);
      fputs ("200 Debug", stdout);
   }
   else
      fputs ("<DNETRAW>", stdout);
   fflush (stdout);

   atexit (&CgiLib__OsuExitRaw);
}  

void CgiLib__OsuExitRaw ()

{
   if (CGILIB_DEBUG) fprintf (stdout, "CgiLib__OsuExitRaw()\n");

   fflush (stdout);
   if (CGILIB_DEBUG)
      fputs ("</DNETTEXT>", stdout);
   else
      fputs ("</DNETRAW>", stdout);
   fflush (stdout);
}  

#endif /* CGILIB_OSU_SUPPORT */

/*****************************************************************************/
/*
Write a single record to the OSU network link, then if expecting a response
wait for one or more records to be returned.  A single line response has no
carriage control.  A multi-line response has each record concatenated into a
single string of newline-separated records.  Returns pointer to static storage!
Open files to the link if first call.
*/

#if CGILIB_OSU_SUPPORT

char* CgiLib__OsuQueryNetLink
(
char *Query,
int ResponseLinesExpected
)
{
   static char  Response [4096];
   static FILE  *NetLinkRead,
                *NetLinkWrite;

   register char  *cptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG)
      fprintf (stdout, "CgiLib__OsuQueryNetLink() |%s|\n", Query);

   if (NetLinkRead == NULL)
   {
      /*************************/
      /* first call, open link */
      /*************************/

      if ((NetLinkRead = fopen ("NET_LINK:", "r", "ctx=rec")) == NULL)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, read open error\n");
         exit (status);
      }
      if ((NetLinkWrite = fopen ("NET_LINK:", "w", "ctx=bin")) == NULL)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, write open error\n");
         exit (status);
      }
   }

   if (Query != NULL)
   {
      /*********/
      /* write */
      /*********/

      if (fputs (Query, NetLinkWrite) == EOF)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, write error\n");
         exit (status);
      }
      fflush (NetLinkWrite);
   }

   if (ResponseLinesExpected > 1)
   {
      /**************************/
      /* multiple line response */
      /**************************/

      cptr = Response;
      for (;;)
      {
         if (fgets (cptr, sizeof(Response)-(cptr-Response), NetLinkRead) == NULL)
         {
            status = vaxc$errno;
            fprintf (stdout, "%%CGILIB-E-NETLINK, read error\n");
            exit (status);
         }
         /** if (CGILIB_DEBUG) fprintf (stdout, "cptr |%s|\n", cptr); **/
         if (*cptr != '\n')
         {
            while (*cptr) cptr++;
            continue;
         }
         *cptr = '\0';
         break;
      }
   }
   else
   if (ResponseLinesExpected > 0)
   {
      /************************/
      /* single line response */
      /************************/

      if (fgets (Response, sizeof(Response), NetLinkRead) == NULL)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, read error\n");
         exit (status);
      }

      for (cptr = Response; *cptr; cptr++);
      if (cptr > Response)
      {
         if (*--cptr == '\n')
            *cptr = '\0';
         else
            cptr++;
      }
   }
   else
      *(cptr = Response) = '\0';

   if (CGILIB_DEBUG) fprintf (stdout, "%d |%s|\n", cptr-Response, Response);
   return (Response);
}  

#endif /* CGILIB_OSU_SUPPORT */

/*****************************************************************************/
/*
Read a POSTed request body into a single array of char (doesn't matter whether
it's text or binary).  Returns NULL if it's not a POST, a pointer if it is. 
The caller must free() this memory when finished with it.  Note: the OSU DECnet
scripting stream is NET_LINK:, the WASD CGI DECnet scripting stream is NET$LINK,
the two should not be confused.
*/

#if CGILIB_POST_SUPPORT

char* CgiLibReadRequestBody
(
char **BodyPtrPtr,
int *BodyLengthPtr
)
{
   static int  DECnetLink = 0,
               InitializeEnvironment = 1,
               NetLinkReadFd = 0,
               OsuEnvironment = 0,
               WasdEnvironment = 0;

   int  status,
        BufferCount,
        BufferSize,
        ContentLength,
        ReadCount;
   char  *cptr,
         *BufferPtr = NULL;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLibReadRequestBody()\n");

   if (InitializeEnvironment)
   {
      InitializeEnvironment = 0;
      WasdEnvironment = (getenv ("HTTP$INPUT") != NULL);
      if (WasdEnvironment)
         DECnetLink = (getenv ("NET$LINK") != NULL);
#if CGILIB_OSU_SUPPORT
      else
         OsuEnvironment = (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL);
#endif

      if (WasdEnvironment)
      {
         if (DECnetLink)
            NetLinkReadFd = open ("NET$LINK:", O_RDONLY, 0, "ctx=bin");
         else
            NetLinkReadFd = open ("HTTP$INPUT:", O_RDONLY, 0, "ctx=bin");
      }
      else
#if CGILIB_OSU_SUPPORT
      if (OsuEnvironment)
         NetLinkReadFd = open ("NET_LINK:", O_RDONLY, 0, "ctx=bin");
      else
#endif
         NetLinkReadFd = open ("SYS$INPUT:", O_RDONLY, 0, "ctx=bin");

      if (NetLinkReadFd == -1)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, binary read open error\n");
         exit (status);
      }
   }

   ContentLength = atoi(CgiLib__GetVar ("CONTENT_LENGTH", ""));
   if (CGILIB_DEBUG) fprintf (stdout, "ContentLength: %d\n", ContentLength);

   if (!ContentLength)
   {
      *BodyPtrPtr = NULL;
      *BodyLengthPtr = 0;
      return (NULL);
   }

   if ((BufferPtr = calloc (1, ContentLength+1)) == NULL)
      exit (vaxc$errno);
   BufferSize = ContentLength;
   BufferCount = 0;

   while (BufferCount < BufferSize)
   {
#if CGILIB_OSU_SUPPORT
      if (OsuEnvironment) CgiLib__OsuQueryNetLink ("<DNETINPUT>", 0);
#endif

      ReadCount = read (NetLinkReadFd,
                        BufferPtr+BufferCount,
                        BufferSize-BufferCount);
      if (CGILIB_DEBUG) fprintf (stdout, "ReadCount: %d\n", ReadCount);
      if (!ReadCount) break;

      if (ReadCount == -1)
      {
         status = vaxc$errno;
         fprintf (stdout, "%%CGILIB-E-NETLINK, binary read error\n");
         exit (status);
      }

      BufferCount += ReadCount;
   }
   if (CGILIB_DEBUG)
      fprintf (stdout, "size: %d count: %d\n", BufferSize, BufferCount);

   /* terminate with a null in case it's text (not counted in 'BufferCount') */
   BufferPtr[BufferCount] = '\0';
   /** if (CGILIB_DEBUG) fprintf (stdout, "|%s|\n", BufferPtr); **/

   *BodyPtrPtr = BufferPtr;
   *BodyLengthPtr = BufferCount;
   if (CGILIB_DEBUG) fprintf (stdout, "%d %d\n", *BodyLengthPtr, *BodyPtrPtr);

   return (BufferPtr);
}

#endif /* CGILIB_POST_SUPPORT */

/*****************************************************************************/
/*
Converts the POSTed url-form-encoded body parameter into "WWW_FORM_field-name"
CGI variables accessable from the variable list.  Field names are limited to
255 characters, after which they are truncated.
*/

#if CGILIB_POST_SUPPORT

int CgiLibFormRequestBody
(
char *BodyPtr,
int BodyLength
)
{
   static char  VarName [256+9] = "WWW_FORM_";

   register char  *cptr, *sptr, *zptr;

   int  VarCount;
   char  *BufferPtr,
         *VarValuePtr;

   /*********/
   /* begin */
   /*********/

   if (CGILIB_DEBUG) fprintf (stdout, "CgiLibFormRequestBody()\n");

   if (!BodyLength) return (0);

   cptr = CgiLib__GetVar ("CONTENT_TYPE", "");
   if (strcmp (cptr, "application/x-www-form-urlencoded")) return (0);

   if ((cptr = BufferPtr = calloc (1, BodyLength+1)) == NULL)
      exit (vaxc$errno);
   memcpy (cptr, BodyPtr, BodyLength);
   cptr[BodyLength] = '\0';

   VarCount = 0;
   while (*cptr)
   {
      zptr = (sptr = VarName) + sizeof(VarName)-1;
      sptr += 9;
      while (*cptr && *cptr != '=' && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr) while (*cptr && *cptr != '=') cptr++;
      if (!*cptr)
      {
         free (BufferPtr);
         return (VarCount);
      }
      *sptr = '\0';
      VarValuePtr = ++cptr;
      while (*cptr && *cptr != '&') cptr++;
      if (*cptr) *cptr++ = '\0';
      if (VarName[9]) CgiLibUrlDecode (VarName+9);
      if (VarValuePtr[0]) CgiLibUrlDecode (VarValuePtr);
      CgiLib__VarList (VarName, VarValuePtr, NULL);
      VarCount++;
   }

   free (BufferPtr);
   return (VarCount);
}

#endif /* CGILIB_POST_SUPPORT */

/*****************************************************************************/

